/*  XMMS - Crnbs-platform multimedia player
 *  Copyright (C) 1998-2001  Peter Alm, Mikael Alm, Olle Hallnas,
 *                           Thomas Nilsson and 4Front Technologies
 *  Copyright (C) 1999-2001  Haavard Kvaalen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "xmmsnbs.h"
#include "libxmms/util.h"
#include <errno.h>
#include <nbs.h>

#define NFRAGS		32

static char *buffer;
static gboolean going, prebuffer, paused, unpause, do_pause, remove_prebuffer;
static gint device_buffer_used, buffer_size, prebuffer_size, blk_size = 4096;
static gint rd_index = 0, wr_index = 0;
static gint output_time_offset = 0;
static guint64 written = 0, output_bytes = 0;
static gint flush;
static gint fragsize, device_buffer_size;
static gchar *device_name;
static pthread_t buffer_thread;
static gboolean realtime, select_works;
static int realchannels;
static NBS *nbs;

static int (*nbs_convert_func)(void **data, int length);

struct format_info {
	union {
		AFormat xmms;
		int nbs;
	} format;
	int frequency;
	int channels;
	int bps;
};


/*
 * The format of the data from the input plugin
 * This will never change during a song. 
 */
struct format_info input;

/*
 * The format we get from the effect plugin.
 * This will be different from input if the effect plugin does
 * some kind of format conversion.
 */
struct format_info effect;

/*
 * The format of the data we actually send to the soundcard.
 * This might be different from effect if we need to resample or do
 * some other format conversion.
 */
struct format_info output;


#if 0
static void nbs_calc_device_buffer_used(void)
{
	audio_buf_info buf_info;
	if(paused)
		device_buffer_used = 0;
	else if (!ioctl(fd, SNDCTL_DSP_GETOSPACE, &buf_info))
		device_buffer_used = (buf_info.fragstotal * buf_info.fragsize) - buf_info.bytes;
}
#endif


static gint nbs_downsample(gpointer ob, guint length, guint speed, guint espeed);

static int nbs_calc_bitrate(int nbs_fmt, int rate, int channels)
{
	int bitrate = rate * channels;

	if (nbs_fmt == NBS_FMT_U16_BE || nbs_fmt == NBS_FMT_U16_LE ||
	    nbs_fmt == NBS_FMT_S16_BE || nbs_fmt == NBS_FMT_S16_LE)
		bitrate *= 2;

	return bitrate;
}

static int nbs_get_format(AFormat fmt)
{
	int format = 0;

	switch (fmt)
	{
		case FMT_U8:
			format = NBS_FMT_U8;
			break;
		case FMT_S8:
			format = NBS_FMT_S8;
			break;
		case FMT_U16_LE:
			format = NBS_FMT_U16_LE;
			break;
		case FMT_U16_BE:
			format = NBS_FMT_U16_BE;
			break;
		case FMT_U16_NE:
#ifdef WORDS_BIGENDIAN
			format = NBS_FMT_U16_BE;
#else
			format = NBS_FMT_U16_LE;
#endif
			break;
		case FMT_S16_LE:
			format = NBS_FMT_S16_LE;
			break;
		case FMT_S16_BE:
			format = NBS_FMT_S16_BE;
			break;
		case FMT_S16_NE:
#ifdef WORDS_BIGENDIAN
			format = NBS_FMT_S16_BE;
#else
			format = NBS_FMT_S16_LE;
#endif
			break;
	}

	return format;
}

static void nbs_setup_format(AFormat fmt, int rate, int nch)
{
	effect.format.xmms = fmt;
	effect.frequency = rate;
	effect.channels = nch;
	effect.bps = nbs_calc_bitrate(nbs_get_format(fmt), rate, nch);

	/* Output is fixed format for NBS */
	output.format.nbs = NBS_FORMAT;
	output.frequency = NBS_BITRATE;
	output.channels = NBS_CHANNELS;


	nbs_set_audio_params();

	output.bps = nbs_calc_bitrate(output.format.nbs, output.frequency,
				      output.channels);
}
	

gint nbs_get_written_time(void)
{
	if (!going)
		return 0;
	return (written * 1000) / effect.bps;
}

gint nbs_get_output_time(void)
{
	guint64 bytes;

	if (!nbs || !going)
		return 0;

	bytes = output_bytes < device_buffer_used ?
		0 : output_bytes - device_buffer_used;

	return output_time_offset + ((bytes * 1000) / output.bps);
}

static int nbs_used(void)
{
	if (realtime)
		return 0;
	else
	{
		if (wr_index >= rd_index)
			return wr_index - rd_index;
		return buffer_size - (rd_index - wr_index);
	}
}

gint nbs_playing(void)
{
	if(!going)
		return 0;
	if (!nbs_used() && (device_buffer_used - (3 * blk_size)) <= 0)
		return FALSE;

	return TRUE;
}

gint nbs_free(void)
{
	if (!realtime)
	{
		if (remove_prebuffer && prebuffer)
		{
			prebuffer = FALSE;
			remove_prebuffer = FALSE;
		}
		if (prebuffer)
			remove_prebuffer = TRUE;

		if (rd_index > wr_index)
			return (rd_index - wr_index) - device_buffer_size - 1;
		return (buffer_size - (wr_index - rd_index)) - device_buffer_size - 1;
	}
	else
		if (paused)
			return 0;
		else
			return 1000000;
}

static inline ssize_t write_all(NBS *nbs, const void *buf, size_t count)
{
	ssize_t done = 0;
	do {
		ssize_t n = nbs_write(nbs, buf, (count - done) / (NBS_CHANNELS * 2));
		if (n == -1)
		{
			if (errno == EINTR)
				continue;
			else
				break;
		}
		done += n * (NBS_CHANNELS * 2);
	} while (count > done);

	return done;
}

static void make_stereo(gpointer dataout, gpointer datain, int length)
{
	/* Signededness doesn't matter for this operation */
	unsigned short *ins = datain, *outs = dataout;
	unsigned char *inc = datain, *outc = dataout;
	int x;
	switch (output.format.nbs)
	{
		case NBS_FMT_S16_BE:
		case NBS_FMT_S16_LE:
		case NBS_FMT_U16_BE:
		case NBS_FMT_U16_LE:
			for (x=0;x<length;x++)
				outs[x] = ins[x >> 1];
			break;
		case NBS_FMT_S8:
		case NBS_FMT_U8:
			for (x=0;x<length<<1;x++)
				outc[x] = inc[x >> 1];
			break;
	}
}

static void nbs_write_audio(gpointer data, int length)
{
	AFormat new_format;
	int new_frequency, new_channels;
	EffectPlugin *ep;
	gpointer data2;
	
	
	new_format = input.format.xmms;
	new_frequency = input.frequency;
	new_channels = input.channels;
	
	ep = get_current_effect_plugin();
	if(effects_enabled() && ep && ep->query_format)
	{
		ep->query_format(&new_format,&new_frequency,&new_channels);
	}
	
	if (new_format != effect.format.xmms ||
	    new_frequency != effect.frequency ||
	    new_channels != effect.channels)
	{
		output_time_offset += (output_bytes * 1000) / output.bps;
		output_bytes = 0;
		nbs_setup_format(new_format, new_frequency, new_channels);
	}
	if (effects_enabled() && ep && ep->mod_samples)
		length = ep->mod_samples(&data, length,
					 input.format.xmms,
					 input.frequency,
					 input.channels);
	if (realtime)
	{
		while (nbs_freespace(nbs) < (length / (NBS_CHANNELS * 2)))
		{
			printf("Stuck sleeping waiting for %d...\n", length);
			xmms_usleep(10000);
		}
	}

	if (nbs_convert_func != NULL)
		length = nbs_convert_func(&data, length);

	if (effect.channels < realchannels) {
		/* Gotta double for stereo-only cards */
		data2 = alloca(length * 2);
		if (!data2) {
			static int warnedalready;
			if (!warnedalready) {
				g_warning("Unable to allocate memory for re-channelization: %s\n", strerror(errno));
				warnedalready++;
			}
		} else {
			make_stereo(data2, data, length);
			data = data2;
			length *= 2;
		}
	}

	if (effect.frequency == output.frequency) {
		if (nbs) {
			output_bytes += nbs_write(nbs, data, length / (NBS_CHANNELS * 2)) * (NBS_CHANNELS * 2);
		} else
			g_warning("Trying to write with no NBS client\n");
	} else {
		output_bytes += nbs_downsample(data, length,
					       effect.frequency,
					       output.frequency);
	}						   
}

static void swap_endian(guint16 *data, int length)
{
	int i;
	for (i = 0; i < length; i += 2, data++)
		*data = GUINT16_SWAP_LE_BE(*data);
}

#define NOT_NATIVE_ENDIAN ((IS_BIG_ENDIAN &&				\
			   (output.format.nbs == NBS_FMT_S16_LE ||		\
			    output.format.nbs == NBS_FMT_U16_LE)) ||	\
			  (!IS_BIG_ENDIAN &&				\
			   (output.format.nbs == NBS_FMT_S16_BE ||		\
			    output.format.nbs == NBS_FMT_U16_BE)))


#define RESAMPLE_STEREO(sample_type)				\
do {								\
	const gint shift = sizeof (sample_type);		\
        gint i, in_samples, out_samples, x, delta;		\
	sample_type *inptr = (sample_type *)ob, *outptr;	\
	guint nlen = (((length >> shift) * espeed) / speed);	\
	if (nlen == 0)						\
		break;						\
	nlen <<= shift;						\
	if (NOT_NATIVE_ENDIAN)					\
		swap_endian(ob, length);			\
	if(nlen > nbuffer_size)					\
	{							\
		nbuffer = g_realloc(nbuffer, nlen);		\
		nbuffer_size = nlen;				\
	}							\
	outptr = (sample_type *)nbuffer;			\
	in_samples = length >> shift;				\
        out_samples = nlen >> shift;				\
	delta = (in_samples << 12) / out_samples;		\
	for (x = 0, i = 0; i < out_samples; i++)		\
	{							\
		gint x1, frac;					\
		x1 = (x >> 12) << 12;				\
		frac = x - x1;					\
		*outptr++ =					\
			(sample_type)				\
			((inptr[(x1 >> 12) << 1] *		\
			  ((1<<12) - frac) +			\
			  inptr[((x1 >> 12) + 1) << 1] *	\
			  frac) >> 12);				\
		*outptr++ =					\
			(sample_type)				\
			((inptr[((x1 >> 12) << 1) + 1] *	\
			  ((1<<12) - frac) +			\
			  inptr[(((x1 >> 12) + 1) << 1) + 1] *	\
			  frac) >> 12);				\
		x += delta;					\
	}							\
	if (NOT_NATIVE_ENDIAN)					\
		swap_endian(nbuffer, nlen);			\
	w = write_all(nbs, nbuffer, nlen);			\
} while (0)


#define RESAMPLE_MONO(sample_type)				\
do {								\
	const gint shift = sizeof (sample_type) - 1;		\
        gint i, x, delta, in_samples, out_samples;		\
	sample_type *inptr = (sample_type *)ob, *outptr;	\
	guint nlen = (((length >> shift) * espeed) / speed);	\
	if (nlen == 0)						\
		break;						\
	nlen <<= shift;						\
	if (NOT_NATIVE_ENDIAN)					\
		swap_endian(ob, length);			\
	if(nlen > nbuffer_size)					\
	{							\
		nbuffer = g_realloc(nbuffer, nlen);		\
		nbuffer_size = nlen;				\
	}							\
	outptr = (sample_type *)nbuffer;			\
	in_samples = length >> shift;				\
        out_samples = nlen >> shift;				\
	delta = ((length >> shift) << 12) / out_samples;	\
	for (x = 0, i = 0; i < out_samples; i++)		\
	{							\
		gint x1, frac;					\
		x1 = (x >> 12) << 12;				\
		frac = x - x1;					\
		*outptr++ =					\
			(sample_type)				\
			((inptr[x1 >> 12] * ((1<<12) - frac) +	\
			  inptr[(x1 >> 12) + 1] * frac) >> 12);	\
		x += delta;					\
	}							\
	if (NOT_NATIVE_ENDIAN)					\
		swap_endian(nbuffer, nlen);			\
	w = write_all(nbs, nbuffer, nlen);			\
} while (0)


static gint nbs_downsample(gpointer ob, guint length, guint speed, guint espeed)
{
	guint w = 0;
	static gpointer nbuffer = NULL;
	static gint nbuffer_size = 0;

	switch (output.format.nbs)
	{
		case NBS_FMT_S16_BE:
		case NBS_FMT_S16_LE:
			if (output.channels == 2)
				RESAMPLE_STEREO(gint16);
			else
				RESAMPLE_MONO(gint16);
			break;
		case NBS_FMT_U16_BE:
		case NBS_FMT_U16_LE:
			if (output.channels == 2)
				RESAMPLE_STEREO(guint16);
			else
				RESAMPLE_MONO(guint16);
			break;
		case NBS_FMT_S8:
			if (output.channels == 2)
				RESAMPLE_STEREO(gint8);
			else
				RESAMPLE_MONO(gint8);
			break;
		case NBS_FMT_U8:
			if (output.channels == 2)
				RESAMPLE_STEREO(guint8);
			else
				RESAMPLE_MONO(guint8);
			break;
	}
	return w;
}

void nbs_xwrite(gpointer ptr, int length)
{
	int cnt, off = 0;

	if (!realtime)
	{
		remove_prebuffer = FALSE;

		written += length;
		while (length > 0)
		{
			cnt = MIN(length, buffer_size - wr_index);
			memcpy(buffer + wr_index, (char *)ptr + off, cnt);
			wr_index = (wr_index + cnt) % buffer_size;
			length -= cnt;
			off += cnt;
		}
	}
	else
	{
		if (paused)
			return;
		nbs_write_audio(ptr, length);
		written += length;
	}
}

void nbs_close(void)
{
	if (!going)
		return;
	going = 0;
	if (!realtime)
		pthread_join(buffer_thread, NULL);
	else
	{
#if 0
		ioctl(fd, SNDCTL_DSP_RESET, 0);
		close(fd);
#endif
		if (nbs) {
			nbs_delstream(nbs);
			nbs = NULL;
		}		
	}
	g_free(device_name);
	nbs_get_convert_buffer(0);
	wr_index = 0;
	rd_index = 0;
}

void nbs_flush(gint time)
{
	if (!realtime)
	{
		flush = time;
		while (flush != -1)
			xmms_usleep(10000);
	}
	else
	{
#if 0
		/* No concept of flushing */
		ioctl(fd, SNDCTL_DSP_RESET, 0);
		close(fd);
		fd = open(device_name, O_WRONLY);
		nbs_set_audio_params();
		output_time_offset = time;
		written = ((guint64)time * input.bps) / 1000;
		output_bytes = 0;
#endif		
	}
}

void nbs_pause(short p)
{
	if (!realtime)
	{
		if (p == TRUE)
			do_pause = TRUE;
		else
			unpause = TRUE;
	}
	else
		paused = p;

}

void *nbs_loop(void *arg)
{
	gint length, cnt;
	fd_set set;
	struct timeval tv;

	while (going)
	{
		if (nbs_used() > prebuffer_size)
			prebuffer = FALSE;
		if (nbs_used() > 0 && !paused && !prebuffer)
		{
			tv.tv_sec = 0;
			tv.tv_usec = 10000;
			FD_ZERO(&set);
			FD_SET(nbs_fd(nbs), &set);
			if(!select_works || (select(nbs_fd(nbs) + 1, NULL, &set, NULL, &tv) > 0))
			{
				length = MIN(blk_size, nbs_used());
				while (length > 0)
				{
					cnt = MIN(length,buffer_size-rd_index);
					nbs_write_audio(buffer + rd_index, cnt);
					rd_index=(rd_index+cnt)%buffer_size;
					length-=cnt;				
				}
			}
		}
		else
			xmms_usleep(10000);
		if (do_pause && !paused)
		{
			do_pause = FALSE;
			paused = TRUE;
#if 0
			/* NBS has no flush */
			/*
			 * We lose some data here that is sent to the
			 * soundcard, but not yet played.  I don't
			 * think this is worth fixing.
			 */
			ioctl(fd, SNDCTL_DSP_RESET, 0);
#endif			
		}
		else if (unpause && paused)
		{
			/* NBS has nothing really to have to do to unpause */
			unpause = FALSE;
#if 0
			close(fd);
			fd = open(device_name, O_WRONLY);
			nbs_set_audio_params();
#endif			
			paused = FALSE;
		}

		if (flush != -1)
		{
			/*
			 * This close and open is a work around of a
			 * bug that exists in some drivers which cause
			 * the driver to get fucked up by a reset
			 */

#if 0
			ioctl(fd, SNDCTL_DSP_RESET, 0);
			close(fd);
			fd = open(device_name, O_WRONLY);
			nbs_set_audio_params();
#endif		
			output_time_offset = flush;
			written = ((guint64)flush * input.bps) / 1000;
			rd_index = wr_index = output_bytes = 0;
			flush = -1;
			prebuffer = TRUE;
		}

	}

#if 0
	/* NBS has no reset */
	ioctl(fd, SNDCTL_DSP_RESET, 0);
#endif
	nbs_delstream(nbs);
	nbs = NULL;	
	g_free(buffer);
	pthread_exit(NULL);
}

void nbs_set_audio_params(void)
{
	int frag, stereo, ret;
	struct timeval tv;
	fd_set set;

#if 0
	/* None of this works with nbs */
	ioctl(fd, SNDCTL_DSP_RESET, 0);
	frag = (NFRAGS << 16) | fragsize;
	ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &frag);
	ioctl(fd, SNDCTL_DSP_SETFMT, &output.format.nbs);
	if (ioctl(fd, SNDCTL_DSP_SETFMT, &output.format.nbs) == -1)
		g_warning("SNDCTL_DSP_SETFMT ioctl failed: %s",
			  strerror(errno));
#endif
	/* Don't store in output.channels, but in another variable */
	realchannels = 2;

	nbs_convert_func =
		nbs_get_convert_func(output.format.nbs,
				     nbs_get_format(effect.format.xmms));

	/*
	 * Stupid hack to find out if the driver support selects, some
	 * drivers won't work properly without a select and some won't
	 * work with a select :/
	 */
	
	tv.tv_sec = 0;
	tv.tv_usec = 50000;
	FD_ZERO(&set);
	FD_SET(nbs_fd(nbs), &set);
	ret = select(nbs_fd(nbs) + 1, NULL, &set, NULL, &tv);
	if (ret > 0)
		select_works = TRUE;
	else
		select_works = FALSE;
}

gint nbs_open(AFormat fmt, gint rate, gint nch)
{
	char host[256] = "dunno";
	if (nbs_cfg.use_alt_audio_device && nbs_cfg.alt_audio_device)
		device_name = g_strdup(nbs_cfg.alt_audio_device);
	else
	{
		if (nbs_cfg.audio_device > 0)
			device_name = g_strdup_printf("%s%d", DEV_DSP, nbs_cfg.audio_device);
		else
			device_name = g_strdup(DEV_DSP);
	}
	gethostname(host, sizeof(host));
	nbs =  nbs_newstream("xmms", host, NBS_FLAG_OVERRIDE);
	if (nbs == NULL)
	{
		g_warning("nbs_open(): Failed to open NBS stream: %s", strerror(errno));
		g_free(device_name);
		return 0;
	}
	nbs_setblocking(nbs, 1);
	nbs_setdebug(nbs, 1);
	nbs_setblocksize(nbs, blk_size);
	if (nbs_connect(nbs))
	{
		g_warning("nbs_open(): Failed to connect NBS stream: %s", strerror(errno));
		nbs_delstream(nbs);
		g_free(device_name);
		return 0;
	}
	
	input.format.xmms = fmt;
	input.frequency = rate;
	input.channels = nch;
	input.bps = nbs_calc_bitrate(nbs_get_format(fmt), rate, nch);

	nbs_setup_format(fmt, rate, nch);
	
	realtime = xmms_check_realtime_priority();
	
	if(!realtime)
	{
		buffer_size = (nbs_cfg.buffer_size * input.bps) / 1000;
		if (buffer_size < 8192)
			buffer_size = 8192;
		prebuffer_size = (buffer_size * nbs_cfg.prebuffer) / 100;
		if (buffer_size - prebuffer_size < 4096)
			prebuffer_size = buffer_size - 4096;

		buffer_size += device_buffer_size;
		buffer = g_malloc0(buffer_size);
	}
	flush = -1;
	prebuffer = TRUE;
	wr_index = rd_index = output_time_offset = written = output_bytes = 0;
	paused = FALSE;
	do_pause = FALSE;
	unpause = FALSE;
	remove_prebuffer = FALSE;

	going = 1;
	if (!realtime)
		pthread_create(&buffer_thread, NULL, nbs_loop, NULL);
	return 1;
}
